from ceci.stage import PipelineStage
from ceci_example.types import HDFFile
import numpy as np
from ceci.errors import *
import pytest

# TODO: test MPI facilities properly with:
# https://github.com/rmjarvis/TreeCorr/blob/releases/4.1/tests/mock_mpi.py


class MockCommunicator:
    def __init__(self, size, rank):
        self.size = size
        self.rank = rank

    def Get_size(self):
        return self.size

    def Get_rank(self):
        return self.rank


def test_construct():
    # This one should work
    class TestStage(PipelineStage):
        name = "test"
        inputs = [("inp1", HDFFile)]
        outputs = []
        config_options = {}

        def run(self):
            pass

    assert PipelineStage.get_stage("test") == TestStage

    # this fails if you execute this file directly as the qualified
    # name is different
    if __name__ != "__main__":
        assert TestStage.get_module().endswith("test_stage")

    s = TestStage({"config": "tests/config.yml", "inp1": "tests/test.hdf5"})

    assert s.rank == 0
    assert s.size == 1
    assert s.is_parallel() == False
    assert s.is_mpi() == False
    assert list(s.split_tasks_by_rank(["1", "2", "3"])) == ["1", "2", "3"]

    r = list(s.data_ranges_by_rank(1000, 100))
    assert r[0] == (0, 100)
    assert r[2] == (200, 300)

    r = list(s.iterate_hdf("inp1", "group1", ["x", "y", "z"], 10))
    for ri in r:
        s, e, ri = ri
        assert len(ri["x"] == 10)
    assert np.all(r[4][2]["z"] == [-80, -82, -84, -86, -88, -90, -92, -94, -96, -98])

    # Fake that we are processor 4/10
    comm = MockCommunicator(10, 4)
    s = TestStage({"config": "tests/config.yml", "inp1": "tests/test.hdf5"}, comm=comm)

    assert s.rank == 4
    assert s.size == 10
    assert s.is_parallel() == True
    assert s.is_mpi() == True
    assert list(s.split_tasks_by_rank("abcdefghijklmnopqrst")) == ["e", "o"]

    r = list(s.data_ranges_by_rank(10000, 100))
    assert r[0] == (400, 500)
    assert r[3] == (3400, 3500)

    r = list(s.iterate_hdf("inp1", "group1", ["x", "y", "z"], 10))
    assert len(r) == 1
    st, e, r = r[0]
    assert st == 40
    assert e == 50
    assert np.all(r["x"] == range(40, 50))

    r = list(s.iterate_hdf("inp1", "group1", ["x", "y", "z"], 10, parallel=False))
    for ri in r:
        st, e, ri = ri
        assert len(ri["x"]) == 10
    assert np.all(r[4][2]["z"] == [-80, -82, -84, -86, -88, -90, -92, -94, -96, -98])


    s = TestStage({"config": "tests/config.yml", "inp1": "tests/test.hdf5"})
    r = list(s.iterate_hdf("inp1", "group1", ["w", "x", "y", "z"], 10, longest=True))
    for i, ri in enumerate(r):
        st, e, ri = ri
        if i < 10:
            assert len(ri["x"]) == 10
            assert np.allclose(ri["x"], np.arange(10*i, 10*(i+1)))
            assert np.allclose(ri["w"], np.arange(10*i, 10*(i+1)))
        else:
            assert len(ri["x"]) == 0
            assert len(ri["w"]) == 10
            assert np.allclose(ri["w"], np.arange(10*i, 10*(i+1)))

    # check error message is raise appropriately if longest is not specified
    s = TestStage({"config": "tests/config.yml", "inp1": "tests/test.hdf5"})
    with pytest.raises(ValueError):
        r = list(s.iterate_hdf("inp1", "group1", ["w", "x", "y", "z"], 10, longest=False))



def test_incomplete():
    class Alpha(PipelineStage):
        pass

    assert "Alpha" in PipelineStage.incomplete_pipeline_stages

    with pytest.raises(IncompleteStage):
        PipelineStage.get_stage("Alpha")


def test_auto_name():
    class Bravo(PipelineStage):
        inputs = []
        outputs = []
        config_options = {}

        def run(self):
            pass

    assert Bravo.name == "Bravo"
    assert PipelineStage.get_stage("Bravo") is Bravo


def test_duplicate():
    class Charlie(PipelineStage):
        inputs = []
        outputs = []

        def run(self):
            pass

    assert PipelineStage.get_stage("Charlie") is Charlie

    with pytest.raises(DuplicateStageName):

        class Charlie(PipelineStage):
            inputs = []
            outputs = []

            def run(self):
                pass

    with pytest.raises(DuplicateStageName):
        # Name it specified and duplicated
        class AlsoCharlie(PipelineStage):
            name = "Charlie"
            inputs = []
            outputs = []

            def run(self):
                pass

    # Should work okay as name is overwritten
    class Charlie(PipelineStage):
        name = "NotCharlie"
        inputs = []
        outputs = []

        def run(self):
            pass

    assert Charlie.name == "NotCharlie"
    assert PipelineStage.get_stage("NotCharlie") is Charlie


def test_explicit_config():
    with pytest.raises(ReservedNameError):

        class Delta(PipelineStage):
            inputs = [("config", None)]
            outputs = []
            config_options = {}

            def run(self):
                pass


def test_okay_abc_dupe_name():
    class Echo(PipelineStage):
        pass

    # okay, as the parent is intended as abstract
    class Echo(Echo):
        inputs = []
        outputs = []

        def run(self):
            pass

    assert Echo.name == "Echo"
    assert PipelineStage.get_stage("Echo") is Echo


def test_okay_abc_dupe_name2():
    class FoxtrotBase(PipelineStage):
        name = "Foxtrot"
        pass

    # okay, as the parent is intended as abstract
    class Foxtrot(FoxtrotBase):
        inputs = []
        outputs = []

        def run(self):
            pass

    assert Foxtrot.name == "Foxtrot"
    assert PipelineStage.get_stage("Foxtrot") is Foxtrot


def test_config_specified():
    # check for incomplete classes
    with pytest.raises(ReservedNameError):

        class Golf(PipelineStage):
            config = "golf"

    # check for complete classes
    with pytest.raises(ReservedNameError):

        class Golf(PipelineStage):
            config = "golf"
            inputs = []
            outputs = []

            def run(self):
                pass

def test_bool_flags():
    class Hotel(PipelineStage):
        inputs = []
        outputs = []
        config_options = {'xyz': bool}

    cmd = ["Hotel", "--config", "tests/config.yml"]

    # Basic case with a single flag
    h = Hotel(Hotel._parse_command_line(cmd + ["--xyz"]))
    assert h.config['xyz'] is True

    h = Hotel(Hotel._parse_command_line(cmd + ["--no-xyz"]))
    assert h.config['xyz'] is False

    # check latter takes precedence if both specified
    h = Hotel(Hotel._parse_command_line(cmd + ["--xyz", "--no-xyz"]))
    assert h.config['xyz'] is False

    h = Hotel(Hotel._parse_command_line(cmd + ["--no-xyz", "--xyz"]))
    assert h.config['xyz'] is True

    # flag is not optional here so must be set
    with pytest.raises(ValueError):
        h = Hotel(Hotel._parse_command_line(cmd))

    class Hotel2(PipelineStage):
        inputs = []
        outputs = []
        config_options = {'xyz': False}

    h = Hotel2(Hotel2._parse_command_line(cmd + ["--xyz"]))
    assert h.config['xyz'] is True

    h = Hotel2(Hotel2._parse_command_line(cmd + ["--no-xyz"]))
    assert h.config['xyz'] is False

    h = Hotel2(Hotel2._parse_command_line(cmd + ["--xyz", "--no-xyz"]))
    assert h.config['xyz'] is False

    h = Hotel2(Hotel2._parse_command_line(cmd + ["--no-xyz", "--xyz"]))
    assert h.config['xyz'] is True

    h = Hotel2(Hotel2._parse_command_line(cmd))
    assert h.config['xyz'] is False

    class Hotel3(PipelineStage):
        inputs = []
        outputs = []
        config_options = {'xyz': True}

    h = Hotel3(Hotel3._parse_command_line(cmd + ["--xyz"]))
    assert h.config['xyz'] is True

    h = Hotel3(Hotel3._parse_command_line(cmd + ["--no-xyz"]))
    assert h.config['xyz'] is False

    h = Hotel3(Hotel3._parse_command_line(cmd + ["--xyz", "--no-xyz"]))
    assert h.config['xyz'] is False

    h = Hotel3(Hotel3._parse_command_line(cmd + ["--no-xyz", "--xyz"]))
    assert h.config['xyz'] is True

    h = Hotel3(Hotel3._parse_command_line(cmd))
    assert h.config['xyz'] is True



def test_unknown_stage():
    with pytest.raises(StageNotFound):
        PipelineStage.get_stage("ThisStageIsDeliberatelyLeftBlank")


# could add more tests here for constructor, but the regression tests here and in TXPipe are
# pretty thorough.

if __name__ == "__main__":
    test_construct()



import os
import sys
import time
import collections
import yaml
import shutil
from .stage import PipelineStage
from . import minirunner
from .sites import get_default_site
from .utils import embolden


class StageExecutionConfig:
    """
    This class stores information describing how an individual job is to be executed,
    for example how many cores it is run on, and where.  It does not store the
    job input or output information or anything like that.

    TODO: Consider attaching this object to the stage itself.

    Attributes
    ----------

    name: str
        The name of the stage
    site: Site object
        (default the global default) The site this stage is run on
    nprocess: int
        (default 1) The number of (usually MPI) processes to use for this task
    nodes: int
        (default 1) The number of nodes the processes should be spread over
    threads_per_process: int
        (default 1) The number of (usually OpenMP) threads to use per process.
    mem_per_process: float
        (defaut 2GB) The amount of memory in GB required for the job
    image: str
        (default is the site default) A docker image name to use for this task
    volume: str
        (default is the site default) Any volume mappings in the form
        /path/on/host:/path/on/container that the job needs
    """

    def __init__(self, info):

        # Core attributes - mandatory
        self.name = info["name"]
        self.site = info.get("site", get_default_site())

        # Parallelism attributes - optional
        self.nprocess = info.get("nprocess", 1)
        self.nodes = info.get("nodes", 1)
        self.threads_per_process = info.get("threads_per_process", 1)  #
        self.mem_per_process = info.get("mem_per_process", 2)

        # Container attributes - optional.
        # There may be a default container for the entire site the
        # stage is run on, in which case use that if it is not overridden.
        self.image = info.get("image", self.site.config.get("image"))
        self.volume = info.get("volume", self.site.config.get("volume"))


class Pipeline:
    """
    The Pipeline base class models the shared information and behaviour
    that pipelines need, no matter which workflow manager runs them.

    This includes constructing the pipeline in the first place from
    stages, checking that it is a valid pipeline, finding overall
    inputs and outputs for the pipeline, and constructing the command
    line for a given stage.

    Sub-classes run the pipeline using a given workflow manager.
    """

    def __init__(self, stages, launcher_config):
        """Construct a pipeline using configuraion information.

        Parameters
        ----------
        stages: list[dict]
            Information used to construct each stage and how it is run
        launcher_config: dict
            Any additional configuration that will be needed by the workflow
            management.  The base class does not use this.
        """
        self.launcher_config = launcher_config

        # These are populated as we add stages below
        self.stage_execution_config = {}
        self.stage_names = []

        # Store the individual stage informaton
        for info in stages:
            self.add_stage(info)

    def add_stage(self, stage_info):
        """Add a stage to the pipeline.

        To begin with this stage is not connected to any others -
        that is determined later.

        The configuration info for this stage must contain at least
        the name of the stage and the name of the site where it is
        to be run.  It can also contain information for the
        StageExecutionConfig above describing parallelism
        and container usage.

        The stage_info can contain the following parameters:
            site: Site object
                The site this stage is run on
            nprocess: int
                (default 1) The number of (usually MPI) processes to use for this task
            nodes: int
                (default 1) The number of nodes the processes should be spread over
            threads_per_process: int
                (default 1) The number of (usually OpenMP) threads to use per process.
            mem_per_process: float
                (defaut 2GB) The amount of memory in GB required for the job
            image: str
                (default is the site default) A docker image name to use for this task
            volume: str
                (default is the site default) Any volume mappings in the form
                /path/on/host:/path/on/container that the job needs

        Parameters
        ----------
        stage_info: dict
            Configuration information for this stage. See docstring for info.


        """
        sec = StageExecutionConfig(stage_info)
        self.stage_execution_config[sec.name] = sec
        self.stage_names.append(sec.name)

    def remove_stage(self, name):
        """Delete a stage from the pipeline

        Parameters
        ----------
        name: str
            The name of the stage to remove.

        """
        self.stage_names.remove(name)
        del self.stage_execution_config[name]

    def find_inputs(self, stage, pipeline_files, run_config):
        return {tag: pipeline_files[tag] for tag, _ in stage.inputs}

    def find_outputs(self, stage, run_config):
        """Get the names of all the outputs a stage will generate.

        The name is determined from the output tag, the type of the file,
        and the directory it will be put into.

        Sub-class pipelines might organise results differently, hence this
        being a method of Pipeline rather than PipelineStage.

        Parameters
        ----------
        stage: PipelineStage class
            This usually takes a class, not an instance, since the inputs/outputs are all
            class level variables

        returns: list[str]
            Complete paths to all outputs of the stage.
        """
        outdir = run_config["output_dir"]
        return {tag: f"{outdir}/{ftype.make_name(tag)}" for tag, ftype in stage.outputs}

    def ordered_stages(self, overall_inputs):
        """Produce a linear ordering for the stages.

        Some stages within the pipeline might be ruunnable in parallel; this
        method does not analyze this, since different workflow managers will
        treat this differently.

        The stages in the pipeline are also checked for consistency, to avoid
        circular pipelines (A->B->C->A) and to ensure that all overall inputs
        needed in the pipeline are supplied from the overall inputs.

        The naive ordering algorithm used is faster when the stages are in
        the correct order to start with.  This won't matter unless you have
        a large number of stages.

        Parameters
        ----------
        overall_inputs: dict{str: str}
            Any inputs that do not need to be generated by the pipeline but are
            instead already supplied at the start.  Mapping is from tag -> path.

        Returns
        -------
        ordered_stages: list[PipelineStage]
            The pipeline stages in an order that can be run.

        """
        stage_names = self.stage_names[:]
        stages = [PipelineStage.get_stage(stage_name) for stage_name in stage_names]

        n = len(stage_names)

        # Check for a pipeline output that is already given as an input
        for stage in stages:
            for tag in stage.output_tags():
                if tag in overall_inputs:
                    raise ValueError(
                        f"Pipeline stage {stage.name} "
                        f"generates output {tag}, but "
                        "it is already an overall input"
                    )
        stage_set = {stage for stage in stage_names}
        if len(stage_set) < len(stages):
            raise ValueError("Some stages are included twice in your pipeline")

        # make a dict mapping each tag to the stages that need it
        # as an input. This is the equivalent of the adjacency matrix
        # in graph-speak
        dependencies = collections.defaultdict(list)
        for stage in stages:
            for tag in stage.input_tags():
                dependencies[tag].append(stage)

        # count the number of inputs required by each stage
        missing_input_counts = {stage: len(stage.inputs) for stage in stages}
        found_inputs = set()
        # record the stages which are receiving overall inputs
        for tag in overall_inputs:
            found_inputs.add(tag)
            for stage in dependencies[tag]:
                missing_input_counts[stage] -= 1

        # find all the stages that are ready because they have no missing inputs
        queue = [stage for stage in stages if missing_input_counts[stage] == 0]
        ordered_stages = []

        # make the ordering
        while queue:
            # get the next stage that has no inputs missing
            stage = queue.pop()
            # for file that stage produces,
            for tag in stage.output_tags():
                # find all the next_stages that depend on that file
                found_inputs.add(tag)
                for next_stage in dependencies[tag]:
                    # record that the next stage now has one less
                    # missing dependency
                    missing_input_counts[next_stage] -= 1
                    # if that stage now has no missing stages
                    # then enqueue it
                    if missing_input_counts[next_stage] == 0:
                        queue.append(next_stage)
            ordered_stages.append(stage)

        # If any stages are still not in the list then there is a problem.
        # Try to diagnose it here.
        if len(ordered_stages) != n:
            stages_missing_inputs = [
                stage for (stage, count) in missing_input_counts.items() if count > 0
            ]
            msg1 = []
            for stage in stages_missing_inputs:
                missing_inputs = [
                    tag for tag in stage.input_tags() if tag not in found_inputs
                ]
                missing_inputs = ", ".join(missing_inputs)
                msg1.append(f"Stage {stage.name} is missing input(s): {missing_inputs}")

            msg1 = "\n".join(msg1)
            raise ValueError(
                f"""
Some required inputs to the pipeline could not be found,
(or possibly your pipeline is cyclic):

{msg1}
"""
            )

        return ordered_stages

    def generate_full_command(
        self, stage, inputs, outputs, config, run_config, missing_inputs_in_outdir=False
    ):
        """Generate the complete command to run a stage.

        This includes any mpirun commands, settings of env vars, and
        if docker/shifter are used it includes that too.  This depends
        on the execution settings (nprocess, nthread) as well as the
        stage itself.

        Parameters
        ----------
        stage: PipelineStage
            The stage to generate the command for

        inputs: dict{str: str}
            Mapping of inputs to the stage to paths

        config: str
            Path to stage configuration file for pipeline stages

        outdir: dict{str: str}
            Mapping of outputs of the stage to paths

        Returns
        -------
        cmd: str
            Complete command to be executed

        """
        core = stage.generate_command(inputs, config, outputs)
        sec = self.stage_execution_config[stage.name]
        cmd = sec.site.command(core, sec)
        return cmd

    def run(self, overall_inputs, run_config, stages_config):
        # Make a copy, since we'll be modifying this.
        pipeline_files = overall_inputs.copy()

        # Get the stages in the order we need.
        stages = self.ordered_stages(overall_inputs)

        # Initiate the run.
        # This is an implementation detail for the different subclasses to store
        # necessary information about the run if necessary.
        # Usually, the arguments are ignored, but they are provided in case a class needs to
        # do something special with any of them.
        run_info = self.initiate_run(stages, pipeline_files, run_config, stages_config)

        # make sure output directories exist
        os.makedirs(run_config["output_dir"], exist_ok=True)
        os.makedirs(run_config["log_dir"], exist_ok=True)

        for stage in stages:
            # If we are in "resume" mode and the pipeline has already been run
            # then we re-use any existing outputs.  User is responsible for making
            # sure they are complete!
            if self.should_skip_stage(stage, run_config):
                self.already_finished_job(stage, run_info)
                output_paths = self.find_outputs(stage, run_config)
                pipeline_files.update(output_paths)

            # Otherwise, run the pipeline and register any outputs from the
            # pipe element.
            else:
                stage_outputs = self.enqueue_job(
                    stage, pipeline_files, stages_config, run_info, run_config
                )
                pipeline_files.update(stage_outputs)

        status = self.run_jobs(run_info, run_config)

        # When the
        self.run_info = run_info
        self.pipeline_outputs = self.find_all_outputs(stages, run_config)

        return status

    def already_finished_job(self, stage, run_info):
        print(f"Skipping stage {stage.name} because its outputs exist already")

    def should_skip_stage(self, stage, run_config):
        outputs = self.find_outputs(stage, run_config).values()
        already_run_stage = all(os.path.exists(output) for output in outputs)
        return already_run_stage and run_config["resume"]

    def find_all_outputs(self, stages, run_config):
        outputs = {}
        for stage in stages:
            stage_outputs = self.find_outputs(stage, run_config)
            outputs.update(stage_outputs)
        return outputs


class DryRunPipeline(Pipeline):
    """A pipeline subclass which just does a dry-run, showing which commands
    would be executed.

    See the base class for almost all behaviour.

    No additional attributes.
    """

    def initiate_run(self, stages, overall_inputs, run_config, stages_config):
        return []

    def should_skip_stage(self, stage, run_config):
        return False

    def enqueue_job(self, stage, pipeline_files, stages_config, run_info, run_config):
        outputs = self.find_outputs(stage, run_config)
        cmd = self.generate_full_command(
            stage, pipeline_files, outputs, stages_config, run_config
        )

        # Replace the first instance of the stage name with bold
        # text, but only if we are printing to screen. This helps the
        # eye pick out the stage you want to run.
        if sys.stdout.isatty():
            cmd = cmd.replace(stage.name, embolden(stage.name), 1)

        run_info.append(cmd)
        return outputs

    def run_jobs(self, run_info, run_config):
        for cmd in run_info:
            print(cmd)
            print("\n")
        return 0

    def find_all_outputs(self, stages, run_config):
        return {}


class ParslPipeline(Pipeline):
    """A subclass of Pipeline that uses Parsl to manage workflow."""

    def initiate_run(self, stages, overall_inputs, run_config, stages_config):
        return []  # list of futures

    def enqueue_job(self, stage, pipeline_files, stages_config, run_info, run_config):
        from parsl.data_provider.files import File

        log_dir = run_config["log_dir"]
        # convert the command into an app
        app = self.generate_app(stage, run_config)

        # Convert the dicts of inputs/outputs to the list that
        # parsl wants.
        # The inputs that exist already need to be converted into Parsl File objects.
        # The ones that don't stay as data futures
        inputs1 = self.find_inputs(stage, pipeline_files, run_config)
        inputs = [
            File(val) if isinstance(val, str) else val for val in inputs1.values()
        ]
        inputs.append(File(stages_config))
        # The outputs are just strings.  python dicts are now ordered,
        # so this works okay.
        outputs = [File(f) for f in self.find_outputs(stage, run_config).values()]

        # have parsl queue the app
        future = app(inputs=inputs, outputs=outputs)
        run_info.append((stage.name, future))
        return {tag: future.outputs[i] for i, tag in enumerate(stage.output_tags())}

    def run_jobs(self, run_info, run_config):
        from parsl.app.errors import BashExitFailure

        log_dir = run_config["log_dir"]
        # Wait for the final results, from all files
        for stage_name, future in run_info:
            try:
                # This waits for b/g pipeline completion.
                future.result()
            # Parsl emits this on any non-zero status code.
            except BashExitFailure:
                stdout_file = f"{log_dir}/{stage_name}.err"
                stderr_file = f"{log_dir}/{stage_name}.out"
                sys.stderr.write(
                    f"""
*************************************************
Error running pipeline stage {stage_name}.

Standard output and error streams below.

*************************************************

Standard output:
----------------

"""
                )
                if os.path.exists(stdout_file):
                    sys.stderr.write(open(stdout_file).read())
                else:
                    sys.stderr.write("STDOUT MISSING!\n\n")

                sys.stderr.write(
                    f"""
*************************************************

Standard error:
----------------

"""
                )

                if os.path.exists(stderr_file):
                    sys.stderr.write(open(stderr_file).read())
                else:
                    sys.stderr.write("STDERR MISSING!\n\n")
                return 1
        return 0

    def generate_app(self, stage, run_config):
        """Build a parsl app that wraps this pipeline stage.

        This object is passed onto parsl for execution.

        Parameters
        ----------
        stage: PipelineStage
            The stage to wrap

        log_dir: str
            The directory to put the logs in.

        sec: StageExecutionConfig
            The execution info for this stage.

        Returns
        -------
        app: App
            A parsl app object

        """
        import parsl

        module = stage.get_module()
        module = module.split(".")[0]

        inputs = {}
        outputs = {}

        # Parsl wants a function with the inputs and outputs
        # extracted from lists.  That function should iself return
        # a string representing the bash cmd line.
        # We build up all these components here.

        # Parsl wants our functions to take their input/output paths
        # from inputs[0], inputs[1], etc.
        for i, inp in enumerate(stage.input_tags()):
            inputs[inp] = f"{{inputs[{i}]}}"
        for i, out in enumerate(stage.output_tags()):
            outputs[out] = f"{{outputs[{i}]}}"

        # The last input file is always the config file
        config_index = len(stage.input_tags())
        config = f"{{inputs[{config_index}]}}"

        # This includes all the "mpirun" stuff.
        sec = self.stage_execution_config[stage.name]
        executor = sec.site.info["executor"]

        # Construct the command line call
        core = self.generate_full_command(stage, inputs, outputs, config, run_config)
        cmd1 = sec.site.command(core, sec)
        log_dir = run_config["log_dir"]

        # We will be exec'ing this here.  We can't just define it inline
        # because it just gets too painful with all the i/o names, so instead
        # we build and exec a string.
        template = f"""
@parsl.app.app.bash_app(executors=[executor])
def {stage.name}(inputs, outputs, stdout='{log_dir}/{stage.name}.out', stderr='{log_dir}/{stage.name}.err'):
    cmd = '{cmd1}'.format(inputs=inputs, outputs=outputs)
    print("Launching command:")
    print(cmd, " 2> {log_dir}/{stage.name}.err 1> {log_dir}/{stage.name}.out")
    return cmd
"""
        print(template)

        # local variables for creating this function.
        d = {"executor": executor.label, "cmd1": cmd1}
        exec(template, {"parsl": parsl}, d)

        # Return the function itself.
        return d[stage.name]


class MiniPipeline(Pipeline):
    """A pipeline subclass that uses Minirunner, a sub-module
    of ceci, to run.

    Minirununer is a small tool I wrote suitable for interactive
    jobs on cori.  It launches jobs locally (not through a batch system),
    which parsl can also do, but it has a simple and clearer (to me at least)
    understanding of available nodes and cores.

    """

    def __init__(self, *args, **kwargs):
        """Create a MiniRunner Pipeline

        In addition to parent initialization parameters (see the
        Pipeline base class), this subclass can take these optional
        keywords.

        Parameters
        ----------
        callback: function(event_type: str, event_info: dict)
            A function called when jobs launch, complete, or fail,
            and when the pipeline aborts.  Can be used for tracing
            execution.  Default=None.

        sleep: function(t: float)
            A function to replace time.sleep called in the pipeline
            to wait until the next time to check process completion
            Most normal usage will not need this.  Default=None.
        """
        self.callback = kwargs.pop("callback", None)
        self.sleep = kwargs.pop("sleep", None)
        super().__init__(*args, **kwargs)

    def build_dag(self, stages, jobs):
        """Build a directed acyclic graph of a set of stages.

        The DAG is represented by a list of jobs that each other job
        depends on.  If all a job's dependencies are complete
        then it can be run.

        Fun fact: the word "dag" is also Australian slang for a
        "dung-caked lock of wool around the hindquarters of a sheep".
        and is used as a semi-affectionate insult.

        Parameters
        ----------

        stages: list[PipelineStage]
            A list of stages to generate the DAG for

        """
        depend = {}
        # for each stage in our pipeline ...
        for stage in stages[:]:
            depend[jobs[stage.name]] = []
            job = jobs[stage.name]
            # check for each of the inputs for that stage ...
            for tag in stage.input_tags():
                for potential_parent in stages[:]:
                    # if that stage is supplied by another pipeline stage
                    if tag in potential_parent.output_tags():
                        depend[job].append(jobs[potential_parent.name])
        return depend

    def initiate_run(self, stages, overall_inputs, run_config, stages_config):
        jobs = {}
        stages = []
        return jobs, stages

    def enqueue_job(self, stage, pipeline_files, stages_config, run_info, run_config):
        jobs, stages = run_info
        sec = self.stage_execution_config[stage.name]
        outputs = self.find_outputs(stage, run_config)
        cmd = self.generate_full_command(
            stage, pipeline_files, outputs, stages_config, run_config
        )
        job = minirunner.Job(
            stage.name,
            cmd,
            cores=sec.threads_per_process * sec.nprocess,
            nodes=sec.nodes,
        )
        jobs[stage.name] = job
        stages.append(stage)
        return outputs

    def run_jobs(self, run_info, run_config):
        jobs, stages = run_info
        # Now the jobs have all been queued, build them into a graph
        graph = self.build_dag(stages, jobs)
        nodes = get_default_site().info["nodes"]
        log_dir = run_config["log_dir"]

        # This pipeline, being mainly for testing, can only
        # run at a single site, so we can assume here that the
        # sites are all the same
        sec = self.stage_execution_config[self.stage_names[0]]
        nodes = sec.site.info["nodes"]

        # Run under minirununer
        runner = minirunner.Runner(
            nodes, graph, log_dir, callback=self.callback, sleep=self.sleep
        )
        interval = self.launcher_config.get("interval", 3)
        try:
            runner.run(interval)
        except minirunner.FailedJob as error:
            sys.stderr.write(
                f"""
*************************************************
Error running pipeline stage {error.job_name}.

Standard output and error streams in {log_dir}/{error.job_name}.out
*************************************************
"""
            )
            return 1

        return 0


class CWLPipeline(Pipeline):
    """Export the pipeline as Common Workflow Language files and optionally run it
    with cwltool or another CWL-aware runner.

    """

    def make_inputs_file(self, stages, overall_inputs, stages_config, inputs_file):

        # find out the class of the file objects.  This is a bit ugly,
        # but the only way we record this now it in the stages.
        input_types = {}
        for stage in stages:
            for tag, ftype in stage.inputs:
                if (tag in overall_inputs) and (tag not in input_types):
                    input_types[tag] = ftype

        inputs = {}
        # First we build up the file inputs, each with this dictionary
        # information.
        for tag, filepath in overall_inputs.items():
            ftype = input_types[tag]
            filepath = os.path.abspath(filepath)
            d = {"class": "File", "path": filepath, "format": ftype.format}
            inputs[tag] = d

        # CWL also wants the config information passed through in an inputs
        # file, so it is all collected together.
        stage_config_data = yaml.safe_load(open(stages_config))
        global_config = stage_config_data.get("global", {})

        # For each stage, we check if any of its config information
        # is set in the config file
        for stage in stages:
            # There might be nothing if no options are needed.
            this_stage_config = stage_config_data.get(stage.name, {})
            # Record only keys that have been set.  If any are missing
            # it is an error that will be noticed later.
            for key in stage.config_options:
                val = this_stage_config.get(key, global_config.get(key))
                if val is not None:
                    inputs[f"{key}@{stage.name}"] = val

        inputs["config"] = {
            "class": "File",
            "path": os.path.abspath(stages_config),
            # YAML file indicator:
            "format": "http://edamontology.org/format_3750",
        }

        # Save to the inputs file
        with open(inputs_file, "w") as f:
            yaml.dump(inputs, f)

    def initiate_run(self, stages, overall_inputs, run_config, stages_config):
        from cwlgen.workflow import Workflow

        wf = Workflow()

        cwl_dir = self.launcher_config["dir"]
        os.makedirs(cwl_dir, exist_ok=True)

        # Write the inputs file
        inputs_file = f"{cwl_dir}/cwl_inputs.yml"
        self.make_inputs_file(stages, overall_inputs, stages_config, inputs_file)

        # CWL treats overall inputs differently, storing
        # them in the inputs file.  We keep the
        overall_inputs.clear()

        return {
            "workflow": wf,
            "cwl_dir": cwl_dir,
            "inputs_file": inputs_file,
            # keeps track of overall pipeline inputs we have already found
            "workflow_inputs": set(),
            "workflow_outputs": {},
        }

    def enqueue_job(self, stage, pipeline_files, stages_config, run_info, run_config):
        from cwlgen.workflowdeps import WorkflowStep, WorkflowStepInput
        from cwlgen.workflowdeps import WorkflowOutputParameter, InputParameter

        cwl_dir = run_info["cwl_dir"]
        workflow = run_info["workflow"]
        log_dir = run_config["log_dir"]

        # Create a CWL representation of this step
        cwl_tool = stage.generate_cwl(log_dir)
        cwl_tool.export(f"{cwl_dir}/{stage.name}.cwl")

        # Load that representation again and add it to the pipeline
        step = WorkflowStep(stage.name, run=f"{cwl_tool.id}.cwl")

        # For CWL these inputs are a mix of file and config inputs,
        # so not he same as the pipeline_files we usually see
        for inp in cwl_tool.inputs:

            # If this input is an putput from an earlier stage
            # then it takes its name based on that
            if inp.id in pipeline_files:
                name = pipeline_files[inp.id] + "/" + inp.id
            # otherwise if it's a config option we mangle
            # it to avod clashes
            elif inp.id in stage.config_options:
                name = f"{inp.id}@{cwl_tool.id}"
            # otherwise just leave it as-is
            else:
                name = inp.id

            # If it's an overall input to the entire pipeline we
            # record that.  We only want things that aren't outputs
            # (first clause) and that we haven't already recorded (second)
            if (inp.id not in pipeline_files) and (
                name not in run_info["workflow_inputs"]
            ):
                run_info["workflow_inputs"].add(name)
                # These are the overall inputs to the enture pipeline.
                # Convert them to CWL input parameters
                cwl_inp = InputParameter(
                    name, label=inp.label, param_type=inp.type, param_format=inp.format
                )
                cwl_inp.default = inp.default

                # Bypassing cwlgen type check in case of arrays
                if type(inp.type) == dict:
                    cwl_inp.type = inp.type

                # record that these are overall pipeline inputs
                workflow.inputs.append(cwl_inp)

            # Record that thisis an input to the step.
            step.inputs.append(WorkflowStepInput(input_id=inp.id, source=name))

        # Also record that we want all the pipeline outputs
        for tag, ftype in stage.outputs:
            # Record the expected output for this tag
            step.out.append(tag)

            # Also record that each file is an output to the entire pipeline
            cwl_out = WorkflowOutputParameter(
                tag,
                f"{step.id}/{tag}",
                label=tag,
                param_type="File",
                param_format=ftype.__name__,
            )
            workflow.outputs.append(cwl_out)

        # Also capture stdout and stderr as outputs
        cwl_out = WorkflowOutputParameter(
            f"{step.id}@stdout",
            output_source=f"{step.id}/{step.id}@stdout",
            label="stdout",
            param_type="File",
        )
        step.out.append(f"{step.id}@stdout")
        workflow.outputs.append(cwl_out)

        cwl_out = WorkflowOutputParameter(
            f"{step.id}@stderr",
            f"{step.id}/{step.id}@stderr",
            label="stderr",
            param_type="File",
        )
        step.out.append(f"{step.id}@stderr")
        workflow.outputs.append(cwl_out)

        # This step is now ready - add it to the list
        workflow.steps.append(step)

        # In CWL our data elemnts dict just records which step each
        # output is made in
        return {tag: step.id for tag in stage.output_tags()}

    def run_jobs(self, run_info, run_config):
        workflow = run_info["workflow"]
        cwl_dir = run_info["cwl_dir"]
        output_dir = run_config["output_dir"]
        log_dir = run_config["log_dir"]
        inputs_file = run_info["inputs_file"]
        workflow.export(f"{cwl_dir}/pipeline.cwl")

        # If 'launcher' is defined, use that
        launcher = self.launcher_config.get(
            "launch",
            f"cwltool --outdir {output_dir} " "--preserve-environment PYTHONPATH",
        )
        if launcher == "cwltool":
            launcher = (
                f"cwltool --outdir {output_dir} " "--preserve-environment PYTHONPATH"
            )

        if launcher:
            # need to include the CWD on the path for CWL as it
            # runs in an isolated directory
            pypath = os.environ.get("PYTHONPATH", "")
            os.environ["PYTHONPATH"] = pypath + ":" + os.getcwd()
            cmd = f"{launcher} {cwl_dir}/pipeline.cwl {inputs_file}"
            print(cmd)
            status = os.system(cmd)
            if pypath:
                os.environ["PYTHONPATH"] = pypath
            else:
                del os.environ["PYTHONPATH"]

        # Parsl insists on putting everything in the same output directory,
        # both logs and file outputs.
        # we need to move those

        if status == 0:
            for step in run_info["workflow"].steps:
                shutil.move(f"{output_dir}/{step.id}.out", f"{log_dir}/{step.id}.out")
                shutil.move(f"{output_dir}/{step.id}.err", f"{log_dir}/{step.id}.err")

        return status



import warnings


class Site:
    """"""

    default_mpi_command = "mpirun -n"

    def __init__(self, config):
        self.mpi_command = config.get("mpi_command", self.default_mpi_command)
        self.info = {}
        self.config = config

    def check_import(self, launcher):
        requirements = {
            "parsl": ["parsl"],
            "cwl": ["cwlgen", "cwltool"],
            "mini": ["psutil"],
        }
        if launcher not in requirements:
            raise ValueError(f"Unknown launcher '{launcher}'")
        missing = []
        libs = requirements[launcher]
        for lib in libs:
            try:
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=DeprecationWarning)
                    __import__(lib)
            except ImportError:
                missing.append(lib)
        if missing:
            missing = ", ".join(missing)
            raise ImportError(
                f"You must install these libraries "
                f"to use the {launcher} launcher: {missing}"
            )

    def configure_for_launcher(self, launcher):
        self.check_import(launcher)
        configure = getattr(self, f"configure_for_{launcher}", None)
        if configure is None:
            raise ValueError(
                f"Site {self} does not know how to configure for launcher {launcher}"
            )
        configure()



from ceci.sites import load, get_default_site
from ceci.pipeline import StageExecutionConfig
from ceci.main import run
import pytest
import os

def test_cori_error():
    # check that errors when trying to run multi-process
    # jobs on cori login nodes are handled correctly.
    # should fail unless dry-run is set.

    launcher_config = {
        "name": "mini",
        "interval": 1.0,
    }
    site_config = {
        "name": "cori-interactive",
    }

    stage_config = {
        "name": "Test",
        "nprocess": 2,
    }

    load(launcher_config, [site_config])
    site = get_default_site()
    sec = StageExecutionConfig(stage_config)

    # should fail if we don't set dry-run
    with pytest.raises(ValueError):
        site.command("xxx", sec)

    # should work if we do set dry-run
    launcher_config["dry_run"] = True
    load(launcher_config, [site_config])
    site = get_default_site()
    site.command("xxx", sec)


def test_in2p3():
    # check that errors when trying to run multi-process
    # jobs on cori login nodes are handled correctly.
    # should fail unless dry-run is set.
    os.environ["NSLOTS"] = "8"

    launcher_config = {
        "name": "mini",
        "interval": 1.0,
    }
    site_config = {
        "name": "cc-parallel",
    }

    stage_config = {
        "name": "Test",
        "nprocess": 2,
    }

    load(launcher_config, [site_config])
    site = get_default_site()
    sec = StageExecutionConfig(stage_config)

    # should work if we do set dry-run
    launcher_config["dry_run"] = True
    load(launcher_config, [site_config])
    site = get_default_site()
    cmd = site.command("xxx", sec)
    assert "singularity" not in cmd

    stage_config["image"] = "imaginary_file.sif"
    sec = StageExecutionConfig(stage_config)


    load(launcher_config, [site_config])
    site = get_default_site()
    cmd = site.command("xxx", sec)
    assert "singularity run" in cmd



from ceci.sites.cori import parse_int_set
from ceci.main import override_config
from ceci.utils import embolden


def test_parse_ints():
    assert parse_int_set("1,2,3") == set([1, 2, 3])
    assert parse_int_set("10-12") == set([10, 11, 12])
    assert parse_int_set("10-12,15,19-21") == set([10, 11, 12, 15, 19, 20, 21])


def test_override_config():
    config = {
        "a": "b",
        "c": {"d": "e"},
        "h": True,
        "i": 8,
        "j": 17.5,
    }
    override_config(config, ["a=a", "c.d=e", "f.x.y.z=g", "h=False", "i=9", "j=19.5"])

    assert config["a"] == "a"
    assert config["c"] == {"d": "e"}
    assert config["f"] == {"x": {"y": {"z": "g"}}}
    assert config["h"] is False
    assert config["i"] == 9
    assert config["j"] == 19.5

def test_embolden():
    x = 'hj6_9xx0'
    y = embolden(x)
    assert x in embolden(x)
    assert y[4:-4] == x



import os
import yaml
import sys
import argparse
import subprocess
from . import pipeline
from .sites import load, set_default_site, get_default_site
from .utils import extra_paths

# Add the current dir to the path - often very useful
sys.path.append(os.getcwd())

parser = argparse.ArgumentParser(
    description="Run a Ceci pipeline from a configuration file"
)
parser.add_argument(
    "pipeline_config", help="Pipeline configuration file in YAML format."
)
parser.add_argument(
    "--dry-run",
    action="store_true",
    help="Just print out the commands the pipeline would run without executing them",
)
parser.add_argument(
    "extra_config",
    nargs="*",
    help="Over-ride the main pipeline yaml file e.g. launcher.name=cwl",
)


def run(pipeline_config_filename, extra_config=None, dry_run=False):
    """
    Runs the pipeline.

    The pipeline takes a main configuration file, which specifies
    which stages are to be run, and how, where to look for them, what overall
    inputs there are to the pipeline, and where to find a file configuring
    the individual stages. See test.yml for an example.

    The extra_config argument lets you override parameters in the config file,
    as long as they are stored in dictionaries.
    For example, if the config file contains:
    launcher:
        name: x

    Then you could include "launcher.name=y" to change this to "y".

    Parameters
    ----------
    pipeline_config_filename: str
        The path to the configuration file
    extra_config: list[str]
        Config parameters to override
    dry_run: bool
        Whether to do a dry-run of the pipeline, not running anything.
    """
    # YAML input file.
    # Load the text and then expand any environment variables
    raw_config_text = open(pipeline_config_filename).read()
    config_text = os.path.expandvars(raw_config_text)
    # Then parse with YAML
    pipe_config = yaml.safe_load(config_text)

    if extra_config:
        override_config(pipe_config, extra_config)

    # parsl execution/launcher configuration information
    launcher_config = pipe_config.get("launcher", {"name": "mini"})
    launcher_name = launcher_config["name"]

    # Launchers may need to know if this is a dry-run
    launcher_config["dry_run"] = dry_run

    # Later we will add these paths to sys.path for running here,
    # but we will also need to pass them to the sites below so that
    # they can be added within any containers or other launchers
    # that we use
    paths = pipe_config.get("python_paths", [])
    if isinstance(paths, str):
        paths = paths.split()

    # Python modules in which to search for pipeline stages
    modules = pipe_config["modules"].split()

    # Required configuration information
    # List of stage names, must be imported somewhere
    stages = pipe_config["stages"]

    # configure the default site based on the config.
    # TODO: allow multiple sites in config files
    # current default site, to be restored later
    default_site = get_default_site()

    site_config = pipe_config.get("site", {"name": "local"})
    # Pass the paths along to the site
    site_config["python_paths"] = paths
    load(launcher_config, [site_config])

    # Inputs and outputs
    inputs = pipe_config["inputs"]
    stages_config = pipe_config["config"]

    # Pre- and post-scripts are run locally
    # before and after the pipeline is complete
    # They are called with the same arguments as
    # this script.  If the pre_script returns non-zero
    # then the pipeline is not run.
    # These scripts ar not run in the dry-run case.
    # In both cases we pass the script the pipeline_config
    # filename and any extra args.
    pre_script = pipe_config.get("pre_script")
    post_script = pipe_config.get("post_script")
    script_args = [pipeline_config_filename]
    if extra_config:
        script_args += extra_config

    run_config = {
        "output_dir": pipe_config["output_dir"],
        "log_dir": pipe_config["log_dir"],
        "resume": pipe_config["resume"],
    }

    # Choice of actual pipeline type to run
    if dry_run:
        pipeline_class = pipeline.DryRunPipeline
    elif launcher_name == "cwl":
        pipeline_class = pipeline.CWLPipeline
    elif launcher_name == "parsl":
        pipeline_class = pipeline.ParslPipeline
    elif launcher_name == "mini":
        pipeline_class = pipeline.MiniPipeline
    else:
        raise ValueError("Unknown pipeline launcher {launcher_name}")

    # temporarily add the paths to sys.path,
    # but remove them at the end
    with extra_paths(paths):

        for module in modules:
            __import__(module)

        # Run the pre-script.  Since it's an error for this to fail (because
        # it is useful as a validation check) then we raise an error if it
        # fails using check_call.
        if pre_script and not dry_run:
            subprocess.check_call(pre_script.split() + script_args, shell=True)

        # Create and run the pipeline
        try:
            p = pipeline_class(stages, launcher_config)
            status = p.run(inputs, run_config, stages_config)
        finally:
            # The load command above changes the default site.
            # So that this function doesn't confuse later things,
            # reset that site now.
            set_default_site(default_site)

        if status:
            return status

        # Run the post-script.  There seems less point raising an actual error
        # here, as the pipeline is complete, so we just issue a warning and
        # return a status code to the caller (e.g. to the command line).
        # Thoughts on this welcome.
        if post_script and not dry_run:
            return_code = subprocess.call(post_script.split() + script_args, shell=True)
            if return_code:
                sys.stderr.write(
                    f"\nWARNING: The post-script command {post_script} "
                    "returned error status {return_code}\n\n"
                )
            return return_code
        # Otherwise everything must have gone fine.
        else:
            return status


def override_config(config, extra):
    print("Over-riding config parameters from command line:")

    for arg in extra:
        key, value = arg.split("=", 1)
        item = key.split(".")
        p = config
        print(f"    {key}: {value}")

        for x in item[:-1]:
            if x in p:
                p = p[x]
            else:
                p[x] = {}
                p = p[x]
        p[item[-1]] = yaml.safe_load(value)


def main():
    args = parser.parse_args()
    status = run(args.pipeline_config, args.extra_config, dry_run=args.dry_run)
    if status == 0:
        print("Pipeline successful.  Joy is sparked.")
    else:
        print("Pipeline failed.  No joy sparked.")
    return status


if __name__ == "__main__":
    sys.exit(main())



"""
Minirunner is a very minimal execution level of a workflow manager.

It understands jobs requiring multiple nodes or cores.
It does minimal checking.

It launches only local jobs, so is designed for debugging or for use on NERSC interactive mode.
"""
import subprocess
import os
import time
import socket
import sys
from timeit import default_timer

# Constant indicators
COMPLETE = 0
WAITING = 1

EVENT_LAUNCH = "launch"
EVENT_ABORT = "abort"
EVENT_FAIL = "fail"
EVENT_COMPLETED = "completed"


class RunnerError(Exception):
    """Base error class."""

    pass


class CannotRun(RunnerError):
    """Error for when no jobs can be run and the pipeline is blocked."""

    pass


class TimeOut(RunnerError):
    """Error for when no jobs can be run and the pipeline is blocked."""

    pass


class FailedJob(RunnerError):
    """Error for when a job has failed."""

    def __init__(self, msg, job_name):
        super().__init__(msg)
        self.job_name = job_name


class Node:
    """Class for nodes available for a job.

    Nodes have an ID and a number of cores.  Can add more capabilities later
    if needed.
    """

    def __init__(self, node_id, cores):
        """Create a node.

        Parameters
        ----------
        node_id: str
            Name for the node

        cores: int
            Number of cores on the node

        Attributes
        ----------

        id: str
            Name for the node

        cores: int
            Number of cores on the node

        is_assigned: bool
            Whether the node is assigned or not to a job.
        """
        self.id = node_id
        self.cores = cores
        self.is_assigned = False

    def __str__(self):
        return f"Node('{self.id}', {self.cores})"

    __repr__ = __str__

    def __hash__(self):
        return hash(self.id)

    def assign(self):
        """Set this node as assigned to a job"""
        self.is_assigned = True

    def free(self):
        """Set this node as no longer assigned to a job"""
        self.is_assigned = False


class Job:
    """Small wrapper for a job to be run by minirunner, incorporating
    the command line and the resources needed to run it.
    """

    def __init__(self, name, cmd, nodes, cores):
        """Create a node.

        Parameters
        ----------
        name: str
            Name for the job

        cmd: str
            Command line to execute for the job

        cores: int
            Number of cores needed for the job

        nodes: int
            Number of nodes needed for the job

        Attributes
        ----------

        name: str
            Name for the node

        cmd: str
            Command line to execute for the job

        cores: int
            Number of cores needed for the job

        nodes: int
            Number of nodes needed for the job
        """
        self.name = name
        self.cmd = cmd
        self.nodes = nodes
        self.cores = cores

    def __str__(self):
        return f"<Job {self.name}>"

    __repr__ = __str__


def null_callback(event_name, event_data):
    pass


class Runner:
    """The main pipeline runner class.

    User is responsible for supplying the graph of jobs to run on it, etc.

    Attributes
    ----------

    nodes: list[Node]
        Nodes available to run on

    job_graph: dict{Job:Job}
        Indicates what jobs are to be run and which other jobs they depend on

    completed_jobs: list[Job]
        Jobs that have finished

    running: list[Job]
        Jobs currently running

    queued_jobs: list[Job]
        Jobs waiting to be run

    log_dir: str
        Dir where the logs are put
    """

    def __init__(self, nodes, job_graph, log_dir, callback=None, sleep=None):
        """Create a Runner

        Parameters
        ----------
        nodes: list[Node]
            Nodes available to run on

        job_graph: dict{Job:Job}
            Indicates what jobs are to be run and which other jobs they depend on

        log_dir: str
            Dir where the logs are put

        callback: function(event_type: str, event_info: dict)
            A function called when jobs launch, complete, or fail,
            and when the pipeline aborts.  Can be used for tracing
            execution. Default=None.

        sleep: function(t: float)
            A function to replace time.sleep called in the pipeline
            to wait until the next time to check process completion
            Most normal usage will not need this. Default=None.
        """
        self.nodes = nodes
        self.job_graph = job_graph
        self.completed_jobs = []
        self.running = []
        self.log_dir = log_dir
        self.queued_jobs = list(job_graph.keys())

        # By default we use an empty callback
        # and the default sleep
        if callback is None:
            callback = null_callback
        if sleep is None:
            sleep = time.sleep
        self.callback = callback
        self.sleep = sleep

    def run(self, interval, timeout=1e300):
        """Launch the pipeline.

        Parameters
        ----------
        interval: int
            How long to wait between each poll of the jobs.

        """
        status = WAITING
        t0 = default_timer()
        try:
            while status == WAITING:
                status = self._update()
                self.sleep(interval)
                t = default_timer() - t0
                if t > timeout:
                    raise TimeOut(
                        f"Pipeline did not finish within {timeout} seconds. "
                        f"These jobs were still running: {self.running}"
                    )
        except Exception:
            # The pipeline should be cleaned up
            # in the event of any error.
            # There should be nothing to clean up
            # if the pipeline ends cleanly
            # TODO: add a test for this
            self.abort()
            raise

    def abort(self):
        """End the pipeline and kill all running jobs."""
        for process, job, alloc in self.running:
            process.kill()

        for node in self.nodes:
            node.free()

        # run the callback, listing all jobs that were running
        # when the system failed.  The failed job will already
        # have triggered EVENT_FAIL.
        self.callback(EVENT_ABORT, {"running": self.running[:]})
        self.running = []

    def _launch(self, job, alloc):
        # launch the specified job on the specified nodes
        # dict alloc maps nodes to numbers of cores to be used
        print(f"\nExecuting {job.name}")
        for node in alloc:
            node.assign()

        cmd = job.cmd

        print(f"Command is:\n{cmd}")
        stdout_file = f"{self.log_dir}/{job.name}.out"
        print(f"Output writing to {stdout_file}\n")

        stdout = open(stdout_file, "w")
        # launch cmd in a subprocess, and keep track in running jobs
        p = subprocess.Popen(cmd, shell=True, stdout=stdout, stderr=subprocess.STDOUT)
        self.running.append((p, job, alloc))
        self.callback(
            EVENT_LAUNCH,
            {"job": job, "stdout": stdout_file, "process": p, "nodes": alloc},
        )

    def _ready_jobs(self):
        # Find jobs ready to be run now
        return [
            job
            for job in self.queued_jobs
            if all(p in self.completed_jobs for p in self.job_graph[job])
        ]

    def _check_impossible(self):
        n_node = len(self.nodes)
        n_core = sum(node.cores for node in self.nodes)

        for job in self.queued_jobs:
            if job.nodes > n_node:
                raise CannotRun(
                    f"Job {job} cannot be run - it needs {job.nodes}"
                    f" nodes but only {n_node} is/are available"
                )
            elif job.cores > n_core:
                raise CannotRun(
                    f"Job {job} cannot be run - it needs {job.cores}"
                    f" cores but only {n_core} is/are available"
                )

    def _update(self):
        # Iterate, checking the status of all jobs and launching any new ones
        self._check_completed()
        self._check_impossible()

        # If all jobs are done, exit

        if len(self.completed_jobs) == len(self.job_graph):
            return COMPLETE

        # Otherwise check what jobs can now run
        ready = self._ready_jobs()

        # and launch them all
        for job in ready:
            # find nodes to run them on
            alloc = self._check_availability(job)
            if alloc is None:
                continue
            # and launch the job
            self.queued_jobs.remove(job)
            self._launch(job, alloc)

        # indicate that we are still running.
        return WAITING

    def _check_completed(self):
        # check if any running jobs have completed
        completed_jobs = []
        continuing_jobs = []
        # loop through all known running ones
        for process, job, alloc in self.running:
            # check status
            status = process.poll()
            # None indicates job is still running
            if status is None:
                continuing_jobs.append((process, job, alloc))
            # status !=0 indicates error in job.
            # kill everything
            elif status:
                print(f"Job {job.name} has failed with status {status}")
                # Call back with info about the failed job and how it ran,
                # then abort the pipeline to stop other things going on.
                # TODO: offer a mode where other non-dependent jobs keep
                # running
                self.callback(
                    EVENT_FAIL,
                    {"job": job, "status": status, "process": process, "nodes": alloc},
                )
                self.abort()
                raise FailedJob(job.cmd, job.name)
            # status==0 indicates sucess in job, so free resources
            else:
                print(f"Job {job.name} has completed successfully!")
                # Call back with info about the successful job and how it ran
                self.callback(
                    EVENT_COMPLETED,
                    {"job": job, "status": 0, "process": process, "nodes": alloc},
                )
                completed_jobs.append(job)
                for node in alloc:
                    node.free()

        self.running = continuing_jobs

        for job in completed_jobs:
            self.completed_jobs.append(job)

    def _check_availability(self, job):
        # check if there are nodes available to run this job
        # Return them if so, or None if not.
        cores_on_node = []
        remaining = job.cores
        alloc = {}

        free_nodes = [node for node in self.nodes if not node.is_assigned]
        if len(free_nodes) >= job.nodes:
            alloc = free_nodes[: job.nodes]
        else:
            alloc = None

        return alloc



from ceci import PipelineStage
from .types import TextFile, YamlFile


class shearMeasurementPipe(PipelineStage):
    """
    This pipeline element is a template for a shape measurement tool
    """

    name = "shearMeasurementPipe"
    inputs = [("DM", TextFile)]
    outputs = [("shear_catalog", TextFile)]
    config_options = {"metacalibration": bool, "apply_flag": bool}

    def run(self):
        # Retrieve configuration:
        my_config = self.config
        print("Here is my configuration :", my_config)

        for inp, _ in self.inputs:
            filename = self.get_input(inp)
            print(f"    shearMeasurementPipe reading from {filename}")
            open(filename)

        for out, _ in self.outputs:
            filename = self.get_output(out)
            print(f"    shearMeasurementPipe writing to {filename}")
            open(filename, "w").write("shearMeasurementPipe was here \n")


class PZEstimationPipe(PipelineStage):
    name = "PZEstimationPipe"
    inputs = [("DM", TextFile), ("fiducial_cosmology", TextFile)]
    outputs = [("photoz_pdfs", TextFile)]

    def run(self):
        for inp, _ in self.inputs:
            filename = self.get_input(inp)
            print(f"    PZEstimationPipe reading from {filename}")
            open(filename)

        for out, _ in self.outputs:
            filename = self.get_output(out)
            print(f"    PZEstimationPipe writing to {filename}")
            open(filename, "w").write("PZEstimationPipe was here \n")


class WLGCRandoms(PipelineStage):
    name = "WLGCRandoms"
    inputs = [("diagnostic_maps", TextFile)]
    outputs = [("random_catalog", TextFile)]

    def run(self):
        for inp, _ in self.inputs:
            filename = self.get_input(inp)
            print(f"    WLGCRandoms reading from {filename}")
            open(filename)

        for out, _ in self.outputs:
            filename = self.get_output(out)
            print(f"    WLGCRandoms writing to {filename}")
            open(filename, "w").write("WLGCRandoms was here \n")


class WLGCSelector(PipelineStage):
    name = "WLGCSelector"
    inputs = [("shear_catalog", TextFile), ("photoz_pdfs", TextFile)]
    outputs = [("tomography_catalog", TextFile)]
    config_options = {"zbin_edges": [float], "ra_range": [-5.0, 5.0]}

    def run(self):
        print(self.config)
        for inp, _ in self.inputs:
            filename = self.get_input(inp)
            print(f"    WLGCSelector reading from {filename}")
            open(filename)

        for out, _ in self.outputs:
            filename = self.get_output(out)
            print(f"    WLGCSelector writing to {filename}")
            open(filename, "w").write("WLGCSelector was here \n")


class SourceSummarizer(PipelineStage):
    name = "SourceSummarizer"
    inputs = [
        ("tomography_catalog", TextFile),
        ("photoz_pdfs", TextFile),
        ("diagnostic_maps", TextFile),
    ]
    outputs = [("source_summary_data", TextFile)]

    def run(self):
        for inp, _ in self.inputs:
            filename = self.get_input(inp)
            print(f"    SourceSummarizer reading from {filename}")
            open(filename)

        for out, _ in self.outputs:
            filename = self.get_output(out)
            print(f"    SourceSummarizer writing to {filename}")
            open(filename, "w").write("SourceSummarizer was here \n")


class SysMapMaker(PipelineStage):
    name = "SysMapMaker"
    inputs = [("DM", TextFile)]
    outputs = [("diagnostic_maps", TextFile)]

    def run(self):
        for inp, _ in self.inputs:
            filename = self.get_input(inp)
            print(f"    SysMapMaker reading from {filename}")
            open(filename)

        for out, _ in self.outputs:
            filename = self.get_output(out)
            print(f"    SysMapMaker writing to {filename}")
            open(filename, "w").write("SysMapMaker was here \n")


class WLGCTwoPoint(PipelineStage):
    name = "WLGCTwoPoint"
    inputs = [
        ("tomography_catalog", TextFile),
        ("shear_catalog", TextFile),
        ("diagnostic_maps", TextFile),
        ("random_catalog", TextFile),
    ]
    outputs = [("twopoint_data", TextFile)]

    def run(self):
        for inp, _ in self.inputs:
            filename = self.get_input(inp)
            print(f"    WLGCTwoPoint reading from {filename}")
            open(filename)

        for out, _ in self.outputs:
            filename = self.get_output(out)
            print(f"    WLGCTwoPoint writing to {filename}")
            open(filename, "w").write("WLGCTwoPoint was here \n")


class WLGCCov(PipelineStage):
    name = "WLGCCov"
    inputs = [
        ("fiducial_cosmology", TextFile),
        ("tomography_catalog", TextFile),
        ("shear_catalog", TextFile),
        ("source_summary_data", TextFile),
        ("diagnostic_maps", TextFile),
    ]
    outputs = [("covariance", TextFile)]

    def rank_filename(self, rank, size):
        filename = self.get_output("covariance")
        if size == 1:
            fname = filename
        else:
            fname = f"{filename}.{rank}"
        return fname

    def run(self):

        # MPI Information
        rank = self.rank
        size = self.size
        comm = self.comm

        for inp, _ in self.inputs:
            filename = self.get_input(inp)
            print(f"    WLGCCov rank {rank}/{size} reading from {filename}")
            open(filename)

        filename = self.get_output("covariance")
        my_filename = self.rank_filename(rank, size)
        print(f"    WLGCCov rank {rank}/{size} writing to {my_filename}")
        open(my_filename, "w").write(f"WLGCCov rank {rank} was here \n")

        #
        if comm:
            comm.Barrier()

        # If needed, concatenate all files
        if rank == 0 and size > 1:
            f = open(filename, "w")
            print(f"Master process concatenating files:")
            for i in range(size):
                fname = self.rank_filename(i, size)
                print(f"   {fname}")
                content = open(fname).read()
                f.write(content)
            f.close()


class WLGCSummaryStatistic(PipelineStage):
    name = "WLGCSummaryStatistic"
    inputs = [
        ("twopoint_data", TextFile),
        ("covariance", TextFile),
        ("source_summary_data", TextFile),
    ]
    outputs = [("wlgc_summary_data", TextFile)]
    parallel = False

    def run(self):
        for inp, _ in self.inputs:
            filename = self.get_input(inp)
            print(f"    WLGCSummaryStatistic reading from {filename}")
            open(filename)

        for out, _ in self.outputs:
            filename = self.get_output(out)
            print(f"    WLGCSummaryStatistic writing to {filename}")
            open(filename, "w").write("WLGCSummaryStatistic was here \n")


if __name__ == "__main__":
    cls = PipelineStage.main()



from .site import Site
import os
import socket
from ..minirunner import Node


class CCParallel(Site):
    """Object representing execution in the local environment, e.g. a laptop."""
    
    def command(self, cmd, sec):
        """Generate a complete command line to be run with the specified execution variables.

        This builds up the command from the core and adds any docker commands, env var settings,
        or mpirun calls.

        Parameters
        ----------
        cmd: str
            The core command to execute.

        sec: StageExecutionConfig
            sec objects contain info on numbers of processes, threads, etc, and container choices

        Returns
        -------
        full_cmd: str
            The complete decorated command to be executed.
        """

        mpi1 = f"{self.mpi_command} {sec.nprocess} "
        mpi2 = f"--mpi" if sec.nprocess > 1 else ""
        volume_flag = f"--bind {sec.volume} " if sec.volume else ""
        paths = self.config.get("python_paths", [])

        # TODO: allow other container types here, like singularity
        if sec.image:
            # If we are setting python paths then we have to modify the executable
            # here.  This is because we need the path to be available right from the
            # start, in case the stage is defined in a module added by these paths.
            # The --env flags in docker/shifter overwrites an env var, and there
            # doesn't seem to be a way to just append to one, so we have to be a bit
            # roundabout to make this work, and invoke bash -c instead.
            bash_start = "bash -c ' cd /opt/TXPipe && "
            bash_end = "'"

            paths_start = "PYTHONPATH=$PYTHONPATH:" + (":".join(paths)) if paths else ""
            return (
                f"{mpi1} "
                f"singularity run "
                f"--env OMP_NUM_THREADS={sec.threads_per_process} "
                f"{volume_flag} "
                f"{sec.image} "
                f"{bash_start} "
                f"{paths_start} "
                f"{cmd} {mpi2} "
                f"{bash_end}"
            )
        else:
            # In the non-container case this is much easier
            paths_env = (
                "PYTHONPATH=" + (":".join(paths)) + ":$PYTHONPATH" if paths else ""
            )
            return (
                f"OMP_NUM_THREADS={sec.threads_per_process} "
                f"{paths_env} "
                f"{mpi1} "
                f"{cmd} {mpi2}"
            )

    def configure_for_parsl(self):
        raise ValueError("Parsl not configured for CC IN2P3 in ceci yet")


    def configure_for_mini(self):
        import psutil

        total_cores = int(os.environ['NSLOTS'])
        cores_per_node = 16  # seems to be the case
        nodes = total_cores // cores_per_node
        last_node_codes = total_cores % cores_per_node

        nodes = [Node(f"Node_{i}", cores_per_node) for i in range(nodes)]

        if last_node_codes:
            i = len(nodes)
            nodes.append(Node(f"Node_{i}", last_node_codes))

        self.info["nodes"] = nodes

    def configure_for_cwl(self):
        pass



from ceci.main import run
from parsl import clear
import tempfile
import os
import pytest
import subprocess


def run1(*config_changes, dry_run=False, expect_fail=False, expect_outputs=True):
    try:
        with tempfile.TemporaryDirectory() as dirname:
            out_dir = os.path.join(dirname, "output")
            log_dir = os.path.join(dirname, "logs")
            config = [f"output_dir={out_dir}", f"log_dir={log_dir}"]
            config += config_changes
            status = run("tests/test.yml", config, dry_run)
            if expect_fail:
                assert status != 0
            else:
                assert status == 0
            if expect_outputs:
                assert os.path.exists(os.path.join(out_dir, "wlgc_summary_data.txt"))
                assert os.path.exists(os.path.join(log_dir, "WLGCSummaryStatistic.out"))

    finally:
        clear()


def test_run_mini():
    run1()


def test_run_dry_run():
    run1(dry_run=True, expect_fail=False, expect_outputs=False)


def test_run_parsl():
    run1("launcher.name=parsl", "launcher.max_threads=3")


def test_run_cwl():
    run1("launcher.name=cwl", "launcher.dir=tests/cwl") == 0


def test_pre_script():
    # use the bash "true" command to simulate a
    # pre-script suceeding
    run1("pre_script='true'")
    # and false to simulate a failure
    with pytest.raises(subprocess.CalledProcessError):
        # error should happen before we get to the asserts, so no expect_fail etc
        run1("pre_script='false'")


def test_post_script():
    # use the bash "true" command to simulate a
    # pre-script suceeding
    run1("post_script='true'")
    # and false to simulate a failure - should not raise an error
    # but should fail.  Outputs should exist.
    run1("post_script='false'", expect_fail=True, expect_outputs=True)


if __name__ == "__main__":
    test_run_dry_run()
    test_run_parsl()
    test_run_mini()
    test_run_cwl()
    test_pre_script()



import time
import psutil
import threading
import datetime


class MemoryMonitor:
    """
    A monitor which reports on memory usage by this process throughout the lifetime of
    a process.

    The monitor is designed to be run in a thread, which is done automatically in the
    start_in_thread method, and will then continue until either the main thread ends
    or the stop method is called from another thread.

    To print out different process information you could use subclass and override the
    log method.
    """

    def __init__(self, interval=30):
        """Create a memory monitor.

        Parameters
        ----------
        interval: float or int
            The interval in seconds between each report.
            Default is 30 seconds
        """
        self.should_continue = True
        self.interval = interval
        self.process = psutil.Process()

    @classmethod
    def start_in_thread(cls, *args, **kwargs):
        """Create a new thread and run the memory monitor in it

        For parameters, see the init method; all arguments sent to this method are
        passed directly to it.

        Returns
        -------
        monitor: MemoryMonitor
            The monitor, already running in its own thread
        """
        monitor = cls(*args, **kwargs)
        thread = threading.Thread(target=monitor._run)
        thread.start()
        return monitor

    def stop(self):
        """Stop the monitor.

        The monitor will complete its current sleep interval and then end.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        self.should_continue = False

    @staticmethod
    def log(p):
        """Print memory usage information to screen

        By default this method prints the p

        Parameters
        ----------
        p: Process
            A psutil process
        """
        mem = p.memory_info()
        # report time since start of process
        dt = datetime.timedelta(seconds=time.time() - p.create_time())

        # Various memory
        rss = mem.rss / 1e9
        vms = mem.vms / 1e9
        avail = psutil.virtual_memory().available / 1e9

        # For now I don't use the python logging mechanism, but
        # at some point should probably switch to that.
        print(
            f"MemoryMonitor Time: {dt}   Physical mem: {rss:.3f} GB   "
            f"Virtual mem: {vms:.3f} GB   "
            f"Available mem: {avail:.1f} GB"
        )

    def _run(self):
        # there are two ways to stop the monitor - it is automatically
        # ended if the main thread completes.  And it can be stopped
        # manually using the stop method.  Check for both these.
        while threading.main_thread().is_alive():
            if not self.should_continue:
                break
            self.log(self.process)
            time.sleep(self.interval)



from ceci import PipelineStage, MiniPipeline, ParslPipeline, Pipeline, DryRunPipeline
from ceci_example.types import TextFile
from ceci.sites import load, reset_default_site
from ceci.utils import extra_paths
import pytest
from parsl import clear
import yaml
import os
import tempfile
import sys

# This one should work
class AAA(PipelineStage):
    name = "AAA"
    inputs = [("b", TextFile)]
    outputs = [("a", TextFile)]
    config_options = {}

    def run(self):
        pass


class BBB(PipelineStage):
    name = "BBB"
    inputs = [("a", TextFile)]
    outputs = [("b", TextFile)]
    config_options = {}

    def run(self):
        pass


class CCC(PipelineStage):
    name = "CCC"
    inputs = [("b", TextFile)]
    outputs = [("c", TextFile)]
    config_options = {}

    def run(self):
        pass


class DDD(PipelineStage):
    name = "DDD"
    inputs = [("b", TextFile), ("c", TextFile)]
    outputs = [("d", TextFile)]
    config_options = {}

    def run(self):
        pass


def test_orderings():

    # TODO: make it so less boilerplate is needed here
    launcher_config = {"interval": 0.5, "name": "mini"}

    A = {"name": "AAA"}
    B = {"name": "BBB"}
    C = {"name": "CCC"}
    D = {"name": "DDD"}

    # This one should work - basic pipeline
    # as long as we supply input 'a'.
    # order should be A then C
    pipeline = Pipeline([C, B], launcher_config)
    order = pipeline.ordered_stages({"a": "a.txt"})
    assert [s.name for s in order] == ["BBB", "CCC"]

    pipeline = Pipeline([C, D, B], launcher_config)
    order = pipeline.ordered_stages({"a": "a.txt"})
    assert [s.name for s in order] == ["BBB", "CCC", "DDD"]

    # Should fail - missing an input, 'a'
    with pytest.raises(ValueError):
        pipeline = Pipeline([D, C, B], launcher_config)
        order = pipeline.ordered_stages({})

    # Should fail - circular
    with pytest.raises(ValueError):
        pipeline = Pipeline([A, B], launcher_config)
        order = pipeline.ordered_stages({})

    # Should fail - one output is supplied as an input
    # with pytest.raises(ValueError):
    with pytest.raises(ValueError):
        pipeline = Pipeline([A], launcher_config)
        order = pipeline.ordered_stages({"a": "a.txt", "b": "b.txt"})

    # Should fail - repeated stage
    with pytest.raises(ValueError):
        pipeline = Pipeline([A, A], launcher_config)
        order = pipeline.ordered_stages({"b": "b.txt"})


class FailingStage(PipelineStage):
    name = "FailingStage"
    inputs = []
    outputs = [("dm", TextFile)]  # exists already as an input
    config_options = {}

    def run(self):
        raise ValueError("This should not run because its outputs exist.")


def _return_value_test_(resume):
    expected_status = 0 if resume else 1
    # Mini pipeline should not run
    launcher_config = {"interval": 0.5, "name": "mini"}
    run_config = {
        "log_dir": "./tests/logs",
        "output_dir": "./tests/inputs",
        "resume": resume,
    }

    pipeline = MiniPipeline([{"name": "FailingStage"}], launcher_config)
    status = pipeline.run({}, run_config, "tests/config.yml")
    assert status == expected_status

    # Parsl pipeline should not run stage either
    launcher_config = {"name": "parsl"}
    site_config = {"name": "local", "max_threads": 1}
    load(launcher_config, [site_config])
    # the above sets the new default to be the parsl-configured site
    pipeline = ParslPipeline([{"name": "FailingStage"}], launcher_config)
    status = pipeline.run({}, run_config, "tests/config.yml")
    assert status == expected_status
    clear()  # clear parsl settings
    reset_default_site()  # reset so default is minirunner again


def test_resume():
    _return_value_test_(True)


def test_fail():
    _return_value_test_(False)


def test_dry_run(mocker):
    import ceci_example

    # override stdout so that it thinks it's a terminal so
    # that we can test the emboldening
    stdout_mock = mocker.patch("ceci.pipeline.sys.stdout")
    stdout_mock.isatty.return_value = True

    config = yaml.safe_load(open("tests/test.yml"))
    launcher_config = {"interval": 0.5, "name": "mini"}

    pipeline = DryRunPipeline(config["stages"], launcher_config)

    run_config = {
        "log_dir": config["log_dir"],
        "output_dir": config["output_dir"],
        "resume": False,
    }

    status = pipeline.run(config["inputs"], run_config, config["config"])

    assert status == 0
    for cmd in pipeline.pipeline_outputs:
        print(f"running {cmd} with os.system")
        status = os.system(cmd)
        assert status == 0


def test_python_paths():
    # make a temp dir
    with tempfile.TemporaryDirectory() as dirname:
        os.mkdir(dirname + "/pretend")

        # create a subdir of that with a module in
        mod_dir = dirname + "/pretend"
        mod_path = mod_dir + "/pretend_module.py"

        # empty module, just to check it imports
        open(mod_path, "w").close()

        assert os.path.exists(mod_path)
        print(os.listdir(mod_dir))

        # create a stage there that uses the submodule
        stage_path = dirname + "/my_stage.py"
        open(stage_path, "w").write(
            """
import ceci
class MyStage(ceci.PipelineStage):
    name = "MyStage"
    inputs = []
    outputs = []
    config_options = {"x": int}
    def run(self):
        import pretend_module
        assert self.config["x"] == 17
"""
        )

        # pipeline admin
        config_path = dirname + "/config.yml"
        open(config_path, "w").write(
            """
MyStage:
    x: 17
            """
        )

        run_config = {
            "log_dir": dirname,
            "output_dir": dirname,
            "resume": False,
            "python_paths": [dirname, mod_dir],
        }

        launcher_config = {"interval": 0.5, "name": "mini"}
        site_config = {"name": "local", "python_paths": [dirname, mod_dir]}
        load(launcher_config, [site_config])

        # note that we don't add the subdir here
        with extra_paths(dirname):
            import my_stage

            print(os.environ["PYTHONPATH"])
            print(sys.path)
            print(os.listdir(dirname))
            pipeline = MiniPipeline([{"name": "MyStage"}], launcher_config)
            status = pipeline.run({}, run_config, config_path)
            log = open(dirname + "/MyStage.out").read()
            print(log)
            assert status == 0


# this has to be here because we test running the pipeline
if __name__ == "__main__":
    PipelineStage.main()



from .stage import PipelineStage
from .pipeline import Pipeline, MiniPipeline, ParslPipeline, DryRunPipeline
from pkg_resources import DistributionNotFound
from pkg_resources import get_distribution

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass



# Things to test
from ceci.minirunner import (
    Runner,
    WAITING,
    COMPLETE,
    Job,
    Node,
    CannotRun,
    TimeOut,
    FailedJob,
)
from ceci.minirunner import EVENT_LAUNCH, EVENT_COMPLETED, EVENT_FAIL, EVENT_ABORT
import time
from .test_helpers import in_temp_dir
from pytest import raises


@in_temp_dir
def test_minirununer_parallel():
    job1 = Job("Job1", "echo start 1; sleep 1; echo end 1", cores=1, nodes=1)
    job2 = Job("Job2", "echo start 2; sleep 1; echo end 2", cores=1, nodes=1)
    job_dependencies = {
        job1: [],
        job2: [],
    }

    # Two nodes with 1 core each
    node1 = Node("node1", 1)
    node2 = Node("node2", 2)
    nodes = [node1, node2]
    r = Runner(nodes, job_dependencies, ".")

    # Both jobs should be ready to run now.  Order is not required
    assert r.queued_jobs in [[job1, job2], [job2, job1]]
    assert r._ready_jobs() in [[job1, job2], [job2, job1]]
    assert len(r._check_availability(job1)) == 1

    # Launch the jobs.  They will all take 1 second, long enough for the tasks below to work
    status = r._update()
    assert status == WAITING

    # Now both jobs should be running, as there are no dependencies
    assert r._ready_jobs() == []
    assert len(r.running) == 2
    # check can assign a node to the job
    assert len(r.queued_jobs) == 0

    # check pipeline finishes okay
    time.sleep(1.1)
    status = r._update()

    assert status == COMPLETE
    assert r.completed_jobs in [[job1, job2], [job2, job1]]


@in_temp_dir
def test_callback_and_sleep():

    events = []

    def callback(event, info):
        events.append((event, info))

    sleeps = []

    def sleep(t):
        sleeps.append(t)
        time.sleep(t)

    job1 = Job("Job1", "echo start 1; sleep 1; echo end 1", cores=1, nodes=1)
    job2 = Job("Job2", "echo start 2; sleep 1; echo end 2", cores=1, nodes=1)
    # This job is designed to fail
    job3 = Job("Job3", "python does_not_exist_for_test.py", cores=1, nodes=1)
    job_dependencies = {
        job1: [],
        job2: [],
        job3: [job1, job2],
    }

    # Two nodes with 1 core each
    node1 = Node("node1", 1)
    node2 = Node("node2", 2)
    nodes = [node1, node2]
    r = Runner(nodes, job_dependencies, ".", callback=callback, sleep=sleep)
    with raises(FailedJob):
        r.run(interval=1)

    assert 1 in sleeps
    for event in [EVENT_LAUNCH, EVENT_COMPLETED, EVENT_FAIL, EVENT_ABORT]:
        # check that each event is fired at least once
        assert [e for e, i in events if e == event]


@in_temp_dir
def test_minirununer_serial():
    job1 = Job("Job1", "echo start 1; sleep 1; echo end 1", cores=1, nodes=1)
    job2 = Job("Job2", "echo start 2; sleep 1; echo end 2", cores=1, nodes=1)
    job_dependencies = {
        job1: [],
        job2: [job1],
    }

    # Two nodes with 1 core each
    node1 = Node("node1", 1)
    node2 = Node("node2", 2)
    nodes = [node1, node2]
    r = Runner(nodes, job_dependencies, ".")

    # Only one job should be ready to run now.  Order is not required
    assert r.queued_jobs in [[job1, job2], [job2, job1]]
    assert r._ready_jobs() == [job1]
    assert len(r._check_availability(job1)) == 1
    assert len(r._check_availability(job2)) == 1

    # Launch the jobs.  They will all take 1 second, long enough for the tasks below to work
    status = r._update()
    assert status == WAITING

    # Now both jobs should be running, as there are no dependencies
    assert r._ready_jobs() == []
    assert len(r.running) == 1
    # check can assign a node to the job
    assert r.queued_jobs == [job2]

    # now should launch the second job as the first is finished
    time.sleep(1.1)
    status = r._update()
    assert status == WAITING

    assert len(r.running) == 1
    assert r.completed_jobs == [job1]
    assert r.queued_jobs == []
    time.sleep(1.1)

    status = r._update()

    assert status == COMPLETE
    assert r.completed_jobs in [[job1, job2], [job2, job1]]
    assert r.running == []
    assert r.queued_jobs == []


@in_temp_dir
def test_timeout():
    node1 = Node("node1", 1)
    nodes = [node1]
    job1 = Job("Job1", "sleep 60", nodes=1, cores=1)
    job_dependencies = {job1: []}
    r = Runner(nodes, job_dependencies, ".")
    with raises(TimeOut):
        r.run(0.5, timeout=1)


@in_temp_dir
def test_cannot_run():
    node1 = Node("node1", 1)
    nodes = [node1]
    job1 = Job("Job1", "echo start 1", nodes=2, cores=1)
    job_dependencies = {job1: []}
    r = Runner(nodes, job_dependencies, ".")

    with raises(CannotRun):
        r.run(0.5, timeout=5)



import os
from ..minirunner import Node

from .site import Site


class CoriSite(Site):
    default_mpi_command = "srun -u -n"

    def command(self, cmd, sec):
        """Generate a complete command line to be run with the specified execution variables.

        This builds up the command from the core and adds any shiftr commands, env var settings,
        or mpirun calls.

        Parameters
        ----------
        cmd: str
            The core command to execute.

        sec: StageExecutionConfig
            sec objects contain info on numbers of processes, threads, etc, and container choices

        Returns
        -------
        full_cmd: str
            The complete decorated command to be executed.
        """

        # on cori we always use srun, even if the command is a single process
        mpi1 = f"{self.mpi_command} {sec.nprocess} --cpus-per-task={sec.threads_per_process}"
        mpi2 = f"--mpi" if sec.nprocess > 1 else ""
        volume_flag = f"-V {sec.volume} " if sec.volume else ""
        paths = self.config.get("python_paths", [])

        if sec.nodes:
            mpi1 += f" --nodes {sec.nodes}"

        if (
            (sec.nprocess > 1)
            and (os.environ.get("SLURM_JOB_ID") is None)
            and (not self.config.get("dry_run"))
        ):
            raise ValueError(
                "You cannot use MPI (by setting nprocess > 1) "
                "on Cori login nodes, only inside jobs."
            )

        if sec.image:
            # If we are setting python paths then we have to modify the executable
            # here.  This is because we need the path to be available right from the
            # start, in case the stage is defined in a module added by these paths.
            # The --env flags in docker/shifter overwrites an env var, and there
            # doesn't seem to be a way to just append to one, so we have to be a bit
            # roundabout to make this work, and invoke bash -c instead.
            paths_start = (
                ("bash -c 'PYTHONPATH=$PYTHONPATH:" + (":".join(paths)))
                if paths
                else ""
            )
            paths_end = "'" if paths else ""
            return (
                f"{mpi1} "
                f"shifter "
                f"--env OMP_NUM_THREADS={sec.threads_per_process} "
                f"{volume_flag} "
                f"--image {sec.image} "
                f"{paths_start} "
                f"{cmd} {mpi2} "
                f"{paths_end} "
            )
        else:
            paths_env = (
                ("PYTHONPATH=" + (":".join(paths)) + ":$PYTHONPATH") if paths else ""
            )
            return (
                # In the non-container case this is much easier
                f"OMP_NUM_THREADS={sec.threads_per_process} "
                f"{paths_env} "
                f"{mpi1} "
                f"{cmd} {mpi2}"
            )

    def configure_for_mini(self):
        # if on local machine, query available cores and mem, make one node
        slurm = os.environ.get("SLURM_JOB_ID")

        if slurm:
            # running a job, either interactive or batch
            # check the environment to find out what nodes we are using
            node_list = os.environ["SLURM_JOB_NODELIST"]
            # parse node list
            if "[" in node_list:
                body, vals = node_list.split("[", 1)
                ints = parse_int_set(vals.strip("]"))
                node_names = [f"{body}{i}" for i in ints]
            else:
                node_names = [node_list]

            # We use "CPU"s here to indicate the number of processes
            # we can run, not the NERSC docs meaning which is the number of threads.
            # On Haswell there are 32 cores with 2 hyper threads each, so this
            # env var reports 64, but we should run 32 processes.
            # On KNL it's 68 cores with 4 threads each, so we should run 68.
            slurm_cpus_on_node = os.environ.get("SLURM_CPUS_ON_NODE")
            if slurm_cpus_on_node == "64":
                cpus_per_node = 32
            elif slurm_cpus_on_node == "272":
                cpus_per_node = 68
            else:
                print("Cannot detect NERSC system - assuming 32 processes per node")
                cpus_per_node = 32

            # collect list.
            nodes = [Node(name, cpus_per_node) for name in node_names]
        else:
            # running on login node
            # use at most 4 procs to avoid annoying people
            nodes = [Node("cori", 4)]

        self.info["nodes"] = nodes

    def configure_for_cwl(self):
        pass


class CoriBatchSite(CoriSite):
    def configure_for_parsl(self):
        from parsl.executors import IPyParallelExecutor
        from parsl.providers import SlurmProvider

        # Get the site details that we need
        cpu_type = self.config.get("cpu_type", "haswell")
        queue = self.config.get("queue", "debug")
        max_slurm_jobs = self.config.get("max_jobs", 2)
        account = self.config.get("account")
        if account is None:
            print(
                "Using LSST DESC account. Specify 'account' in the site config to override"
            )
            account = "m1727"
        walltime = self.config.get("walltime", "00:30:00")
        setup_script = self.config.get(
            "setup",
            "/global/projecta/projectdirs/lsst/groups/WL/users/zuntz/setup-cori",
        )

        provider = SlurmProvider(
            partition=queue,  # Replace with partition name
            min_blocks=0,  # one slurm job to start with
            max_blocks=max_slurm_jobs,  # one slurm job to start with
            scheduler_options=f"#SBATCH --constraint={cpu_type}\n"
            f"#SBATCH --account={account}\n"
            f"#SBATCH --walltime={walltime}\n",
            nodes_per_block=1,
            init_blocks=1,
            worker_init=f"source {setup_script}",
        )

        executor = IPyParallelExecutor(
            label="cori-batch",
            provider=provider,
        )

        self.info["executor"] = executor


class CoriInteractiveSite(CoriSite):
    def configure_for_parsl(self):
        from parsl.executors import ThreadPoolExecutor

        max_threads = int(os.environ.get("SLURM_JOB_NUM_NODES", 1))
        executor = ThreadPoolExecutor(label="local", max_threads=max_threads)
        self.info["executor"] = executor


def parse_int_set(nputstr):
    # https://stackoverflow.com/questions/712460/interpreting-number-ranges-in-python/712483
    selection = set()
    invalid = set()
    # tokens are comma seperated values
    tokens = [x.strip() for x in nputstr.split(",")]
    for i in tokens:
        try:
            # typically tokens are plain old integers
            selection.add(int(i))
        except:
            # if not, then it might be a range
            try:
                token = [int(k.strip()) for k in i.split("-")]
                if len(token) > 1:
                    token.sort()
                    # we have items seperated by a dash
                    # try to build a valid range
                    first = token[0]
                    last = token[len(token) - 1]
                    for x in range(first, last + 1):
                        selection.add(x)
            except:
                # not an int and not a range...
                invalid.add(i)
    # Report invalid tokens before returning valid selection
    if invalid:
        raise ValueError(f"Invalid node list: {nputstr}")
    return selection



#!/usr/bin/env python
"""
Lightweight pipeline engine for LSST DESC
Copyright (c) 2018-2020 LSST DESC
http://opensource.org/licenses/MIT
"""
from setuptools import setup
from io import open

# read the contents of the README file
with open('README.md', encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='ceci',
    description='Lightweight pipeline engine for LSST DESC',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/LSSTDESC/ceci',
    maintainer='Joe Zuntz',
    license='BSD',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Development Status :: 3 - Alpha',
    ],
    packages=['ceci', 'ceci.sites'],
    entry_points={
        'console_scripts':['ceci=ceci.main:main']
    },
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    install_requires=['pyyaml', 'psutil'],
    extras_require={
      'parsl': ['flask', 'parsl>=1.0.0'],
      'cwl': ['cwlgen>=0.4', 'cwltool>=2.0.20200126090152'],
      'test': ['pytest', 'codecov', 'pytest-cov'],
      }
)



import time
from ceci.monitor import MemoryMonitor
# should already be installed on CI as we need HDF5
import numpy as np


def test_monitor_report(capsys):
    monitor = MemoryMonitor.start_in_thread(interval=1)
    time.sleep(2)
    x = np.random.uniform(size=10_000_000)
    time.sleep(2)
    monitor.stop()
    captured = capsys.readouterr()
    lines = [line for line in captured.out.split("\n") if line.startswith("MemoryMonitor Time")]
    assert len(lines) >= 3

    # check that memory usage is large enough
    gb1 = float(lines[0].split()[5])
    gb2 = float(lines[-1].split()[5])
    # avoid a rounding error by removing an extra 1%
    assert gb2 - gb1 >= x.nbytes / 1.01e9


from .cori import CoriBatchSite, CoriInteractiveSite
from .local import LocalSite, Site
from .ccin2p3 import CCParallel
import os


# Maps names from config files to classes.
site_classes = {
    "local": LocalSite,
    "cori-interactive": CoriInteractiveSite,
    "cori-batch": CoriBatchSite,
    "cc-parallel": CCParallel,
}


# by default use a local site configured.
# Overwritten if you call load below.


def set_default_site(site):
    global _default_site
    _default_site = site
    return _default_site


def reset_default_site():
    site = LocalSite({"max_threads": 2})
    site.configure_for_mini()
    set_default_site(site)


def get_default_site():
    return _default_site


reset_default_site()


def load(launcher_config, site_configs):
    """Configure a launcher and the sites it will execute on.

    This is the only site function needed in the main entry points to the code.

    The launcher_config argument should contain at least the "name" argument.
    Which additional arguments are accepted then depends on the choice of that.
    See test.yml for examples.

    The site_configs list has a similar structure.

    Parameters
    ----------
    launcher_config: dict
        Configuration information on launchers (parsl, minirunner, CWL).

    site_configs: list[dict]
        list of configs for different sites (local, cori-batch, cori-interactive).

    """
    sites = []

    launcher_name = launcher_config["name"]
    dry_run = launcher_config.get("dry_run", False)

    # Create an object for each site.
    for site_config in site_configs:
        site_name = site_config["name"]
        # Also tell the sites whether this is a dry-run.
        # for example, the cori site checks you're not
        # trying to run srun on a login node, but we skip
        # that test if we are not actually running the command,
        # just printing it.
        site_config["dry_run"] = dry_run

        try:
            cls = site_classes[site_name]
        except KeyError:
            raise ValueError(f"Unknown site {site_name}")

        site = cls(site_config)
        site.configure_for_launcher(launcher_name)
        sites.append(site)

    setup_launcher(launcher_config, sites)

    # replace the default site with the first
    # one found here
    set_default_site(sites[0])

    return sites


def setup_launcher(launcher_config, sites):
    """
    Some launchers need an initial setup function to be run.
    Do that here.
    """
    name = launcher_config["name"]

    if name == "parsl":
        setup_parsl(launcher_config, sites)
    # no setup to do for other managers
    else:
        pass


def setup_parsl(launcher_config, sites):
    """
    Set up parsl for use with the specified sites.
    """
    from parsl import load as parsl_load
    from parsl.config import Config
    from parsl import set_file_logger

    executors = [site.info["executor"] for site in sites]
    config = Config(executors=executors)
    parsl_load(config)

    # Optional logging of pipeline infrastructure to file.
    log_file = launcher_config.get("log")
    if log_file:
        log_file_dir = os.path.split(os.path.abspath(log_file))[0]
        os.makedirs(log_file_dir, exist_ok=True)
        set_file_logger(log_file)
